/* 
 * Copyright (C) 2004 Romain Dolbeau <romain@dolbeau.org>
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License. See the file COPYING in the main directory of this archive for
 *  more details.
 */

package org.dolbeau.mencgen;


import java.util.*;
import java.io.*;

class lavcMencoderWrapperLocal implements lavcMencoderWrapper {
    private String mencoder = "mencoder";
    private String mplayer = "mplayer";

    lavcMencoderWrapperLocal() {
    }

    lavcMencoderWrapperLocal(String me, String mp) {
        mencoder = me;
        mplayer = mp;
    }

    static private double extractPSNRlavc(String line, String key) {
        // PSNR: Y:43.92, Cb:47.75, Cr:48.59, All:44.93
        double result = -1.;
        
        int start = line.indexOf(key);
        if (start != -1) {
            start += (key.length() + 1);
            int stop = line.indexOf(",", start);
            
            if (stop == -1)
                result = Double.parseDouble(line.substring(start));
            else
                result = Double.parseDouble(line.substring(start, stop));
        }
        return result;
    }

    static private double extractPSNRxvid(String line, String key) {
        //xvid: Average PSNR y : 47.47 dB, u : 49.73 dB, v : 50.70 dB, for 47 frames
        double result = -1.;
        int start = line.indexOf(key + " :");
        
        if (start != -1) {
            start += (key.length() + 3);
            int stop = line.indexOf("dB", start);
            result = Double.parseDouble(line.substring(start, stop -1));
        }
        return result;
    }

    static private double extractPSNRx264(String line, String key) {
        //x264 [info]: PSNR Mean Y:43.30 U:47.56 V:49.11 Avg:44.40 Global:43.84 kb/s:554.4 fps:29.434
        double result = -1.;
        int start = line.indexOf(key + ":");
        
        if (start != -1) {
            start += (key.length() + 1);
            //int stop = line.indexOf("dB", start);
            result = Double.parseDouble(line.substring(start, start + 5));
        }
        return result;
    }

    static private double extractPosixTime(String line) {
        // user 0.00
        double result = -1.;

        int start = line.indexOf(" ");
        if (start != -1) {
            start++;
            result = Double.parseDouble(line.substring(start));
        }

        return result;
    }

    static private int extractFromPos(String line) {
        // Pos:  56.0s   2804f ( 0%) 762fps Trem:   7min 2526mb  A-V:0.053 [2617:448] A/Vms 24/0 D/B/S 0/0/1403
        int stop = line.indexOf("f");
        int start = stop - 1;
        while(line.charAt(start) != ' ')
            start--;
        start++;
        return (Integer.parseInt(line.substring(start,stop)));
    }

    public static final int THREAD_COPY = 1;
    public static final int THREAD_CLEAN = 2;
    public static final int THREAD_MAKE = 3;
    public static final int THREAD_RUN = 4;

    public static final int THREAD_STDOUT = 0;
    public static final int THREAD_STDERR = 1;

    protected WatchReaderThread createWatchReaderThread(BufferedReader b, int threadtype, int threadout) {
        return new WatchReaderThread(b, threadtype, threadout);
    }

    protected class WatchReaderThread extends Thread {
        protected final BufferedReader buf;
        protected int threadtype;
        protected int threadout;
        double Y = -1., Cb = -1., Cr = -1., All = -1.;
        double seconds = -1.;
        
        WatchReaderThread(BufferedReader b, int tt, int to) {
            buf = b;
            threadtype = tt;
            threadout = to;
        }
        
        public void run() {
            try {
                String line = null;
                line = buf.readLine();
                while (line != null) {
                    if (line.indexOf("PSNR:") > -1) { // libavcodec
                        lavcGeneticMain.err_println(" -- " + line);
                        Y = extractPSNRlavc(line, "Y");
                        Cb = extractPSNRlavc(line, "Cb");
                        Cr = extractPSNRlavc(line, "Cr");
                        All = extractPSNRlavc(line, "All");
                    }
                    if ((line.indexOf("xvid:") > -1) &&
                        (line.indexOf("Average PSNR") > -1)) { // xvid
                        lavcGeneticMain.err_println(" -- " + line);
                        Y = extractPSNRxvid(line, "y");
                        Cb = extractPSNRxvid(line, "u");
                        Cr = extractPSNRxvid(line, "v");
                    
                        double dY = Math.exp((Y * Math.log(10.)) / -10.);
                        double dCb = Math.exp((Cb * Math.log(10.)) / -10.);
                        double dCr = Math.exp((Cr * Math.log(10.)) / -10.);
                        double dAll = (dY + dCb/4 + dCr/4) / 1.5;
                    
                        All = -10. * Math.log(dAll) / Math.log(10.);
                    }
                    if ((line.indexOf("x264") > -1) &&
                        (line.indexOf("PSNR") > -1) &&
                        (line.indexOf("slice") == -1)) { // x264
                        lavcGeneticMain.err_println(" -- " + line);
                        Y = extractPSNRx264(line, "Y");
                        Cb = extractPSNRx264(line, "U");
                        Cr = extractPSNRx264(line, "V");
                        double Allorig = extractPSNRx264(line, "Avg");

                        double dY = Math.exp((Y * Math.log(10.)) / -10.);
                        double dCb = Math.exp((Cb * Math.log(10.)) / -10.);
                        double dCr = Math.exp((Cr * Math.log(10.)) / -10.);
                        double dAll = (dY + dCb/4 + dCr/4) / 1.5;

                        All = -10. * Math.log(dAll) / Math.log(10.);

                        lavcGeneticMain.err_println("x264: " + Allorig + ", computed: " + All);
                    
                    }
                    if (line.indexOf("user") > -1) {
                        lavcGeneticMain.err_println(" -- " + line);
                        seconds = extractPosixTime(line);
                    }

                    line = buf.readLine();
                }
            } catch (Exception e) { System.err.println("Exception " + e + " while watching a BufferedReader"); }
        }
    }
    
    public lavcResults dumpMoviePassX(lavcEncodeData led , int pass, String machine) {
        double Y = -1., Cb = -1., Cr = -1., All = -1.;
        double seconds = -1.;
        
        final String cmd =
            "/usr/bin/time -p" +
            " " + mencoder +
            " " + led.extraOpts +
            " " + led.sourceFile +
            " -aid " + led.aid + 
            " " + led.videoFilter + 
            " -oac " + led.audioCodec + " " + led.audioOptions +
            " -ovc " + led.codecData.codec + " " + led.codecData.codecOpts + " " + led.getLacvOptsString() + ":" +
            led.codecData.bitrate + "=" + led.bitrate +
            (pass > 0 ? ":" + led.codecData.pass + "=" + pass : "") +
            " -o " + led.output;
        
        lavcGeneticMain.err_println("EXEC (" + machine + "): [pass " + pass + "] " + cmd);
        File workdir = null;
        try {
            workdir = new File(led.workdir);
            if (!workdir.exists()) {
                workdir.mkdirs();
            } else {
                if (pass < 2) {
                    File divx2passlog = new File(led.workdir + "/" + "divx2pass.log");
                    if (divx2passlog.exists())
                        divx2passlog.delete();
                    File xvidtwopassstats = new File(led.workdir + "/" + "xvid-twopass.stats");
                    if (xvidtwopassstats.exists())
                        xvidtwopassstats.delete();
                }
            }
        } catch (Exception e) { lavcGeneticMain.err_println("Exception " + e + " while creating File for " + led.workdir); }
        try {
            final Process theMenc = Runtime.getRuntime().exec(cmd, null, workdir);
            final Thread ourThread = Thread.currentThread();
            final InputStream theMencIn = theMenc.getInputStream();
            final InputStream theMencEr = theMenc.getErrorStream();
            final BufferedReader bufEr = new BufferedReader(new InputStreamReader(theMencEr));
            final BufferedReader bufIn = new BufferedReader(new InputStreamReader(theMencIn));
            final String machineCopy = machine;
            
            WatchReaderThread wrtEr = createWatchReaderThread(bufEr, THREAD_CLEAN, THREAD_STDERR);
            WatchReaderThread wrtIn = createWatchReaderThread(bufIn, THREAD_CLEAN, THREAD_STDOUT);
            
            wrtEr.start();
            wrtIn.start();

            wrtEr.join();
            wrtIn.join();

            All = (wrtEr.All > wrtIn.All) ? wrtEr.All : wrtIn.All;
            Cb = (wrtEr.Cb > wrtIn.Cb) ? wrtEr.Cb : wrtIn.Cb;
            Cr = (wrtEr.Cr > wrtIn.Cr) ? wrtEr.Cr : wrtIn.Cr;
            Y = (wrtEr.Y > wrtIn.Y) ? wrtEr.Y : wrtIn.Y;

            seconds = (wrtEr.seconds > wrtIn.seconds) ? wrtEr.seconds : wrtIn.seconds;

            int result = theMenc.waitFor();
            
            if (result != 0) {
                lavcGeneticMain.err_println("Command: \"" + cmd + "\" returned " + result);
            }
            
            bufIn.close();
            bufEr.close();
            theMencIn.close();
            theMencEr.close();
        } catch (Exception e) { lavcGeneticMain.err_println("Exception " + e + " while processing command \"" + cmd + "\""); }
        
        return new lavcResults(Y, Cb, Cr, All, seconds);
    }
}
