/* 
 * Copyright (C) 2004 Romain Dolbeau <romain@dolbeau.org>
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License. See the file COPYING in the main directory of this archive for
 *  more details.
 */

package org.dolbeau.xsCOEj;

import java.util.*;

/**
 * The abstract Class representing an option of the command
 * @author Romain Dolbeau <romain@dolbeau.org>
 * @version  $Id: Option.java,v 1.6 2005/02/09 14:42:39 dolbeau Exp $
 * @see OptionInstance
 * @see OptionInstanceList
 */
public abstract class Option implements java.io.Serializable, Comparable {
    /** all options */
    private static Set all = new HashSet();
    /** Must occur before instance of those options */
    protected Set before = new HashSet();
    /** Must occur after instance of those options */
    protected Set after = new HashSet();;
    /** Get a List of all possible OptionInstance of that Option
     * @return a List of all possible OptionInstance of that Option
     */
    public abstract List getAllInstances();
    /** Get the description of that option
     * @return a String describing that option
     */
    public abstract String getDescription();
    /** Get the comment of that option
     * @return a String commenting on that option
     */
    public abstract String getComment();
    /** Get a single-instance version of that option
     * @param oi the instance to use
     * @return the same Option with a single instance
     */
    public abstract Option getOneInstanceOption(OptionInstance oi);

    /** index of the instance representing the default value for the command */
    private int defaultCommandIndex = 0;
    /** index of the instance representing the default selected value.
     * -1 means explore all
     * @see OptionChooser
     */
    private int defaultSelectedIndex = -1;
    /** set if the user must choose an instance (no exploration) */
    protected boolean mChoose = false;
    public boolean mustChoose() {
        return mChoose;
    }
    /** set if the option instance must <b>not</b> appear in the command line */
    protected boolean virtualOpt = false;
    public boolean isVirtual() {
        return virtualOpt;
    }

    protected Option() {
        all.add(this);
    }

    public String toString() {
        return "Option " +
            (getComment() ==  null ? "" : "(" + getComment() + ") ") +
            getDescription() + ": " + getAllInstances();
    }
    
    public int getDefaultCommandIndex() { return defaultCommandIndex; };

    public void setDefaultCommandIndex(int index) {
        if ((index < 0) ||
            (index >= getAllInstances().size())) {
            throw new java.lang.IndexOutOfBoundsException("[Option::setDefaultCommandIndex] Not that many (" + index + ") instance in option " + this);
        }
        defaultCommandIndex = index;
    }

    public void setDefaultCommandIndex(OptionInstance loi) {
        setDefaultCommandIndex(getAllInstances().indexOf(loi));
    }

    public int getDefaultSelectedIndex() { return defaultSelectedIndex; };

    public void setDefaultSelectedIndex(int index) {
        if ((index < -1) ||
            (index >= getAllInstances().size())) {
            throw new java.lang.IndexOutOfBoundsException("[Option::setDefaultSelectedIndex] Not that many (" + index + ") instance in option " + this);
        }
        defaultSelectedIndex = index;
    }

    public void setDefaultSelectedIndex(OptionInstance loi) {
        setDefaultSelectedIndex(getAllInstances().indexOf(loi));
    }

    public OptionInstance getDefaultInstance() {
        return (OptionInstance)getAllInstances().get(getDefaultCommandIndex());
    }

    public boolean hasInstance(OptionInstance oi) {
        Iterator it = getAllInstances().iterator();
        while (it.hasNext()) {
            OptionInstance oi2 = (OptionInstance)it.next();
            if (oi2.equals(oi))
                return true;
        }
        return false;
    }

    public boolean equals(Object o) {
        if (o instanceof Option) {
            Option o2 = (Option)o;

            if (getDescription().equals(o2.getDescription()))
                return true;
        }
        return false;
    }

    public int compareTo(Object o) {
        Option o2 = (Option)o;
        int result = 0;
        Iterator it;
        it = before.iterator();
        while (it.hasNext() && (result == 0)) {
            Option ob = (Option)it.next();
            if (ob.equals(o2))
                result = -1;
        }
        it = after.iterator();
        while (it.hasNext() && (result == 0)) {
            Option ob = (Option)it.next();
            if (ob.equals(o2))
                result = 1;
        }
        if (result == 0)
            result = getDescription().compareTo(o2.getDescription());
        return result;
    }
    
    /** Breed two instances of the Option. The resulting OptionInstance will either one of the
     * parent or a random OptionInstance, with a probability in parameter.
     * @param loi1 the first parent
     * @param loi2 the second parent
     * @param oddsR percentage of a random OptionInstance
     * @param random the Random object to use
     * @return an OptionInstance from that Option, breeded from the parent and a random factor.
     */
    public OptionInstance breed(OptionInstance loi1, OptionInstance loi2, int oddsR, Random random) {
        List all = getAllInstances();
        if (loi1.getParentOption() != loi2.getParentOption()) {
            throw new IllegalArgumentException("Can't breed different options (" +
                                               loi1.getParentOption() + " / " +
                                               loi2.getParentOption() + ")");
        }
        if (!all.contains(loi1) ||
            !all.contains(loi2)) {
            throw new IllegalArgumentException("Instance(s) not in option (" +
                                               loi1 + " / " +
                                               loi2 + " in " +
                                               loi1.getParentOption() + ")");
        }

        if (random.nextInt(100) < oddsR) {
            return (OptionInstance)all.get(random.nextInt(all.size()));
        } else {
            if (random.nextInt(100) < 50)
                return loi1;
            else 
                return loi2;
        }
    }

    public int countCompatibleInstance(OptionInstance oi) {
        Iterator it = getAllInstances().iterator();
        int count = 0;
        while (it.hasNext()) {
            OptionInstance loi = (OptionInstance)it.next();
            if (loi.isCompatible(oi))
                count ++;
        }
        return count;
    }
    
    public OptionInstance firstCompatibleInstance(OptionInstance oi) {
        Iterator it = getAllInstances().iterator();
        int count = 0;
        while (it.hasNext()) {
            OptionInstance loi = (OptionInstance)it.next();
            if (loi.isCompatible(oi))
                return loi;
        }
        return null;
    }

    public boolean canUseAverage() {
        return false;
    }

    protected boolean useAverage = false;
    public void setUseAverage(boolean b) {
        useAverage = b;
    }

    public void setAverageParameter(Object o) {
        // do nothing 
    }

    /** Get a List of all possible OptionInstance of that Option except the one in parameter
     * @param oi The OptionInstance to exclude form the returned List
     * @return A List of all possible OptionInstance of that Option, except the one in parameter
     */
    List getAllInstancesBut(OptionInstance oi) {
        List temp = new ArrayList();
        Iterator it = getAllInstances().iterator();
        while (it.hasNext()) {
            OptionInstance oi2 = (OptionInstance)it.next();
            if (!oi2.equals(oi))
                temp.add(oi2);
        }
        return temp;
    }

    void addBefore(Option o) {
        this.before.add(o);
        o.after.add(this);
        closureBeforeAfter();
    }

    private void addBeforeRecur(Option o) {
        if (!this.before.contains(o)) {
            this.before.add(o);
        }
        Iterator it = o.before.iterator();
        while (it.hasNext()) {
            Option o2 = (Option)it.next();
            this.addBeforeRecur(o2);
        }   
        
    }
    
    private void addAfterRecur(Option o) {
        if (!this.after.contains(o)) {
            this.after.add(o);
        }
        Iterator it = o.after.iterator();
        while (it.hasNext()) {
            Option o2 = (Option)it.next();
            this.addAfterRecur(o2);
        }
        
    }
    
    private static void closureBeforeAfter() {
        boolean done = false;
        while (!done) {
            done = true;
            Iterator it = all.iterator();
            while (it.hasNext()) {
                Option o = (Option)it.next();
                Iterator ita = o.after.iterator();
                while (ita.hasNext()) {
                    Option o2 = (Option)ita.next();
                    if (!o2.before.contains(o)) {
                        done = false;
                    }
                    o2.addBeforeRecur(o);
                    
                }
                Iterator itb = o.before.iterator();
                while (itb.hasNext()) {
                    Option o2 = (Option)itb.next();
                    if (!o2.after.contains(o)) {
                        done = false;
                    }
                    o2.addAfterRecur(o);
                }
            }
        }
    }
    
    static Option findByDescription(String n) {
        Iterator it = all.iterator();
        while (it.hasNext()) {
            Option o = (Option)it.next();
            if (o.getDescription().equals(n))
                return o;
        }
        return null;
    }

    OptionInstance findInstanceByString(String n) {
        Iterator it = getAllInstances().iterator();
        while (it.hasNext()) {
            OptionInstance oi = (OptionInstance)it.next();
            if (n == null) {
                if (oi.getOptionString() == null)
                    return oi;
            } else {
                if (n.equals(oi.getOptionString()))
                    return oi;
            }
        }
        return null;
    }

    Set getBefore() {
        return before;
    }
    Set getAfter() {
        return after;
    }
}
