/* 
 * Copyright (C) 2004 Romain Dolbeau <romain@dolbeau.org>
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License. See the file COPYING in the main directory of this archive for
 *  more details.
 */

package org.dolbeau.xsCOEj;

import java.util.*;
import javax.swing.*;
import java.awt.event.*;

/**
 * A swing component to choose between instances of an Option.
 * @author Romain Dolbeau <romain@dolbeau.org>
 * @version  $Id: OptionChooser.java,v 1.2 2004/12/06 14:14:27 dolbeau Exp $
 * @see Option
 * @see OptionInstance
 */
public class OptionChooser extends JPanel implements Comparable {
    private static final String kTryAll = "(try all)";
    private static final String kUnset = "(not set)";
    private final Option myOption;
    private final ButtonGroup bg = new ButtonGroup();
    private final List allButtons = new ArrayList();
    private final Set allChoosers;

    private int index = -1;

    public boolean equals(Object o) {
        if (o instanceof OptionChooser) {
            OptionChooser o2 = (OptionChooser)o;
            return myOption.equals(o2.myOption);
        }
        return false;
    }
    
    public int compareTo(Object o) {
        OptionChooser second = (OptionChooser)o;
        
        if (myOption.mustChoose() && !second.myOption.mustChoose())
            return -1;
        if (!myOption.mustChoose() && second.myOption.mustChoose())
            return 1;

        int i = myOption.getClass().getName().compareTo(second.myOption.getClass().getName());
        if (i != 0)
            return i;

        return myOption.compareTo(second.myOption);
    }
    
    public Option getOption() {
        return myOption;
    }

    public void setIndex(int i) {
        index = i;
        System.err.println("Setting " + myOption.getDescription() +
                           (myOption.getComment() ==  null ? "" : " (" +  myOption.getComment() + ")") +
                           " to " + i);
    }

    class OptionChooserButtonAction extends AbstractAction {
        OptionChooserButtonAction(String name) {
            super(name);
        }
        public void actionPerformed(ActionEvent e) {
            String cmd = e.getActionCommand();
            if (cmd.equals(kTryAll))
                setIndex(-1);

            ListIterator it = myOption.getAllInstances().listIterator();
            while (it.hasNext()) {
                String name = ((OptionInstance)it.next()).toString();
                if ((cmd.equals(name)) ||
                    (cmd.equals(kUnset) && (name == null)))
                    setIndex(it.previousIndex());
            }

            if (index != -1) {
                OptionInstance oi = (OptionInstance)myOption.getAllInstances().get(index);
                Iterator it2;
                it2 = allChoosers.iterator();
                while (it2.hasNext()) {
                    OptionChooser oc = (OptionChooser)it2.next();
                    oc.checkCompatibility(oi);
                }
            }
        }
    }
    
    public OptionChooser(Option m, Set ac) {
        myOption = m;
        allChoosers = ac;
        JRadioButton jrb;

        JComponent localBox = this;
        this.setLayout(new java.awt.FlowLayout(java.awt.FlowLayout.LEADING));
        //this.setMaximumSize(new java.awt.Dimension(800, 800));

        index = myOption.getDefaultSelectedIndex();
        int mEncIndex = myOption.getDefaultCommandIndex();

        //localBox.add(new JLabel(myOption.getDescription() + ": "));
        localBox.setBorder(new javax.swing.border.TitledBorder((myOption.getDescription() +
                                                                (myOption.getComment() ==  null ?
                                                                 "" :
                                                                 " (" +  myOption.getComment() + ")"))));

        JCheckBox avgButton = new JCheckBox(new AbstractAction("Average Values") {
                public void actionPerformed(ActionEvent evt) {
                    JCheckBox jb = (JCheckBox)evt.getSource();
                    myOption.setUseAverage(jb.isSelected());
                }
            });
        avgButton.setEnabled(myOption.canUseAverage());
        localBox.add(avgButton);

        jrb = new JRadioButton(new OptionChooserButtonAction(kTryAll));
        localBox.add(jrb);
        bg.add(jrb);
        allButtons.add(jrb);
        if (m.mustChoose())
            jrb.setEnabled(false);
        if (index == -1)
            jrb.setSelected(true);

        int i = 0;
        Iterator it = myOption.getAllInstances().iterator();
        while (it.hasNext()) {
            OptionInstance loi = (OptionInstance)it.next();
            String name = loi.toString();
            String displayName = (name == null ? kUnset : name);
            jrb = new JRadioButton(new OptionChooserButtonAction(displayName));
            localBox.add(jrb);
            bg.add(jrb);
            allButtons.add(jrb);
            if (index == i)
                jrb.setSelected(true);
            if (mEncIndex == i)
                jrb.setForeground(java.awt.Color.red);
            i++;
        }
        
        if (index != -1) {
            OptionInstance oi = (OptionInstance)myOption.getAllInstances().get(index);
            Iterator it2;
            it2 = allChoosers.iterator();
            while (it2.hasNext()) {
                OptionChooser oc = (OptionChooser)it2.next();
                oc.checkCompatibility(oi);
            }
        }
    }

    private void checkCompatibility(OptionInstance oi) {
        if (index >= 0)
            if (((OptionInstance)myOption.getAllInstances().get(index)).isCompatible(oi))
                return;
        
        int count = myOption.countCompatibleInstance(oi);

        if ((count > 1) &&
            (index == -1))
            return;

        if (count > 1) {
            ((JRadioButton)allButtons.get(0)).setSelected(true);
            index = -1;
        } else {
            int index2 = myOption.getAllInstances().indexOf(myOption.firstCompatibleInstance(oi));
            ((JRadioButton)allButtons.get(index2 + 1)).setSelected(true);
            index = index2;
        }
    }
    
    public Option getProcessedOption() {
        if (index == -1)
            return myOption;

        return myOption.getOneInstanceOption((OptionInstance)myOption.getAllInstances().get(index));
    }
}
